/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2006 NoMachine, http://www.nomachine.com/.               */
/*                                                                        */
/* NXCLIENT, NX protocol compression and NX extensions to this software   */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for applicability.       */
/*                                                                        */
/* NX and NoMachine are trademarks of Medialogic S.p.A.                   */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#include <Logger.h>
#include "Service.h"

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <unistd.h>

#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>

#ifndef __CYGWIN32__
#include <netinet/tcp.h>
#include <sys/un.h>
#endif

#ifdef __APPLE__
typedef int socklen_t;
#endif

#define DEBUG

Service::Service(const char* filepath, const char *type)
{
  strncpy(type_, type, DEFAULT_STRING_LENGTH);
  setupPaths(filepath);
  pid_ = 0;
  state_ = STOPPED;
  port_ = 0;
  argumentsNumber_ = 0;
}

Service::~Service()
{
}

int Service::setupPaths(const char* filepath)
{
  memset(arguments_, '\0', MAX_ARGUMENTS * sizeof(char *));
  memset(executablePath_, '\0', DEFAULT_STRING_LENGTH);
  memset(rootPath_, '\0', DEFAULT_STRING_LENGTH);
  memset(typePath_, '\0', DEFAULT_STRING_LENGTH);
  memset(optionsPath_, '\0', DEFAULT_STRING_LENGTH);
  memset(sessionPath_, '\0', DEFAULT_STRING_LENGTH);
  memset(pidPath_, '\0', DEFAULT_STRING_LENGTH);
  memset(commandPath_, '\0', DEFAULT_STRING_LENGTH);
  memset(portPath_, '\0', DEFAULT_STRING_LENGTH);

  #ifdef DEBUG
  logUser("Service::setupPaths filepath='%s'", filepath);
  #endif

  strcpy(rootPath_, filepath);

  strncpy(typePath_, rootPath_, DEFAULT_STRING_LENGTH);
  strncat(typePath_, "/type", DEFAULT_STRING_LENGTH - strlen(rootPath_) -1);

  strncpy(optionsPath_, rootPath_, DEFAULT_STRING_LENGTH);
  strncat(optionsPath_, "/options", DEFAULT_STRING_LENGTH - strlen(rootPath_) -1);

  strncpy(sessionPath_, rootPath_, DEFAULT_STRING_LENGTH);
  strncat(sessionPath_, "/session", DEFAULT_STRING_LENGTH - strlen(rootPath_) -1);

  strncpy(pidPath_, rootPath_, DEFAULT_STRING_LENGTH);
  strncat(pidPath_, "/pid", DEFAULT_STRING_LENGTH - strlen(rootPath_) -1);

  strncpy(commandPath_, rootPath_, DEFAULT_STRING_LENGTH);
  strncat(commandPath_, "/command", DEFAULT_STRING_LENGTH - strlen(rootPath_) -1);

  strncpy(portPath_, rootPath_, DEFAULT_STRING_LENGTH);
  strncat(portPath_, "/port", DEFAULT_STRING_LENGTH - strlen(rootPath_) -1);

  #ifdef DEBUG
  logUser("Service::setupPaths rootPath='%s'", rootPath_);
  #endif

  return 0;
}


bool Service::isRunning()
{
  if(/*state_ == STOPPED ||*/ pid_ == 0)
  {
    return false;
  }
  else
  {
    /*
     * if state is running we need to be sure that
     * it didn't die in the meanwhile
     */
    if (kill(pid_, SIGCONT) == 0)
    {
      return true;
    }
    else
    {
      state_ = STOPPED;
      /* pid_ = 0; */
      return false;
    }
  }
}

int Service::start()
{
  return 0;
}

unsigned int Service::getFreePort(unsigned int startingPort)
{
  int fd = 0;

  sockaddr_in address;

  memset(&address, '\0', sizeof(address));

  if (startingPort <= 0)
  {
    startingPort = 1025;
  }

  fd = socket(PF_INET, SOCK_STREAM, PF_UNSPEC);

  address.sin_family = AF_INET;

  address.sin_addr.s_addr = htonl(INADDR_ANY);


  for (unsigned int port = startingPort; port < 65000; port++)
  {
    address.sin_port = htons(port);

    if (bind(fd, (sockaddr *)&address, sizeof(address)) == 0)
    {
      //
      // it means that everything is ok
      //
      close(fd);
      return port;
    }
  }

  close(fd);

  //
  // in case of problems
  //
  return 0;
}

int Service::writePidFile()
{
  FILE *filePtr = fopen(pidPath_, "w");

  if (filePtr == NULL)
  {
    return -1;
  }

  fprintf(filePtr, "%d", pid_);
  fclose(filePtr);

  return 0;
}

int Service::writePortFile()
{
  if (port_ != 0)
  {
    FILE *filePtr = fopen(portPath_, "w");

    if (filePtr == NULL)
    {
      return -1;
    }

    fprintf(filePtr, "%d", port_);
    fclose(filePtr);
  }
  return 0;
}


int Service::writeCommandFile()
{
  FILE *filePtr = fopen(commandPath_, "w");

  if (filePtr == NULL)
  {
    return -1;
  }

  int n = 0;

  while (arguments_[n])
  {
    fprintf(filePtr, "%s", arguments_[n++]);

    if (arguments_[n])
    {
      fprintf(filePtr, "%s"," ");
    }
  }
  fclose(filePtr);
  return 0;
}


int Service::writeOptionsFile()
{
  FILE *filePtr = fopen(optionsPath_, "w");

  if (filePtr == NULL)
  {
    return -1;
  }

  //
  // we don't want to write the program name
  //
  int n = 1;

  while (arguments_[n])
  {
    bool paramName = false;
    char *str = arguments_[n++];

    if (*str == '-')
    {
      paramName = true;
      str++;
    }
    else
    {
      paramName = false;
    }

    //
    // in case of double dash
    //
    if (*str == '-')
    {
      str++;
    }

    if (!*str)
    {
      continue;
    }

    fprintf(filePtr, "%s", str);

    if (arguments_[n])
    {
      if (*arguments_[n] == '-')
      {
        if (paramName)
        {
          fprintf(filePtr, "%s", "=1,");
        }
        else
        {
          fprintf(filePtr, "%s", ",");
        }
      }
      else
      {
        fprintf(filePtr, "%s", "=");
      }
    }
  }

  fclose(filePtr);

  return 0;
}

int Service::writeTypeFile()
{
  FILE *filePtr = fopen(typePath_, "w");

  if (filePtr == NULL)
  {
    return -1;
  }

  fprintf(filePtr, "%s", type_);
  fclose(filePtr);

  return 0;
}

int Service::run()
{
  writeCommandFile();
  writeOptionsFile();

  logUser("Info: Starting daemon: '%s'", getID());

  if ((pid_ = fork()) != 0)
  {
    if (pid_ < 0)
    {
      logUser("Error: Fork failed.");
    }
    else
    {
      #ifdef DEBUG
      logUser("Info: Created process with pid=%d", pid_);
      #endif
    }

    writePidFile();

    return pid_;
  }

  #ifdef DEBUG
  logUser("Info: Executing child with pid=%d and ppid=%d", getpid(), getppid());
  #endif

  if (execvp(executablePath_, arguments_) == -1)
  {
    logUser("Error: Cannot execute '%s'. Error: '%s'", executablePath_, strerror(errno));
    exit(-1);
  }

  exit(0);

  //
  // just to shut up the compiler
  //
  return 0;
}

int Service::load()
{

  //
  // read file type
  //
  FILE *filePtr = fopen(typePath_, "r");

  if (filePtr == NULL)
  {
    return -1;
  }

  fscanf(filePtr, "%s", type_);
  fclose(filePtr);

  //
  // read pid file
  //

  filePtr = fopen(pidPath_, "r");

  if (filePtr == NULL)
  {
    return -1;
  }

  int pid = 0;

  fscanf(filePtr, "%d", &pid);
  pid_ = pid;
  fclose(filePtr);

  return 0;
}

int Service::getPid()
{
  return pid_;
}

const char* Service::getType()
{
  return type_;
}

int Service::stop()
{
  int ret = kill(pid_, SIGTERM);

  if (!ret)
  {
    renameDirectory("T-");
  }
  else
  {
    renameDirectory("F-");
  }

  return ret;
}

int Service::renameDirectory(const char* prefix)
{
  char *slash = NULL;

  char path[DEFAULT_STRING_LENGTH];

  strcpy(path, rootPath_);

  slash = strrchr(path, '/');

  if (slash == NULL)
  {
    return -1;
  }

  char *name = new char[strlen(slash) + strlen(prefix) + 1];

  if (name == NULL)
  {
    return -1;
  }

  strcpy(name, prefix);

  //
  // this is safe in the worst case slash + 1 will be '\0'
  //
  strcat(name, slash + 1);

  #ifdef DEBUG
  logUser("Info: renameDirectory: name='%s'\n",name);
  #endif

  if (strlen(path) + strlen(prefix) > DEFAULT_STRING_LENGTH)
  {
    return -1;
  }

  strcpy(slash + 1, name);

  if (rename((const char*)rootPath_, (const char*)path))
  {
    logUser("Error: Cannot rename '%s' to '%s'. Error is '%s'", rootPath_, path, strerror(errno));
    return -1;
  }
  else
  {
    return -1;
  }

  setupPaths(path);

  delete [] name;

  return 0;
}

int Service::renameTerminated()
{
  return renameDirectory("T-");
}

int Service::renameFailed()
{
  return renameDirectory("F-");
}

void Service::addArgument(char *arg)
{
   if (argumentsNumber_ < MAX_ARGUMENTS)
   {
      logUser("Service::addArgument[%d][%s]", argumentsNumber_, arg);
      arguments_[argumentsNumber_++] = arg;
   }
   else
   {
      logUser("Service::addArgument: Error cannot add: [%d][%s]", argumentsNumber_, arg);
   }
}

const char* Service::getID()
{
  char *id = strrchr(rootPath_, '-');
  if (id)
	id++;
  return id;
}
